#!/usr/bin/python
#-----------------------------------------------------------------------
# File    : hac.py
# Contents: hierarchical agglomerative clustering
# Author  : Christian Borgelt
# History : 2016.02.24 file created
#-----------------------------------------------------------------------
from sys  import argv, stderr
from time import time
from math import sqrt

#-----------------------------------------------------------------------
# Constants
#-----------------------------------------------------------------------
oo = float('inf')

#-----------------------------------------------------------------------
# Main Program
#-----------------------------------------------------------------------

def error (msg):
    stderr.write(msg); exit()   # print an error message and abort

#-----------------------------------------------------------------------

if __name__ == '__main__':
    name = argv[1] if len(argv) > 1 else 'iris'
    show = argv[2] if len(argv) > 2 else None

    # --- read data ---
    with open(name+'.dom', 'r') as inp:
        atts = [line.rstrip('\r\n').split(')')[0][4:]
                for line in inp if line[:3] == 'dom']
    with open(name+'.tab', 'r') as inp:
        hdr  = inp.readline().rstrip('\n\r').split()
        data = [line.rstrip('\n\r').split() for line in inp]
    cmap = dict(zip(hdr, range(len(hdr))))
    cids = [cmap[a] for a in atts]
    data = [[float(d[i]) for i in cids] for d in data]

    # for d in data: print d      # print the data points

    # --- normalize data ---
    dsum = [0.0 for x in data[0]]
    dsqr = [0.0 for x in data[0]]
    dims = range(len(dsum))     # get the number of dimensions
    for d in data:              # traverse the data points
        for i in dims:          # traverse the dimensions
            dsum[i] += d[i]     # compute sum and sum of squares
            dsqr[i] += d[i]*d[i]
    n = float(len(data))        # get the number of data points
    for i in dims:              # traverse the dimensions
        dsum[i] /= n            # compute mean and standard deviation
        dsqr[i] = sqrt((dsqr[i] -n*dsum[i]*dsum[i]) /n)
        #dsqr[i] = sqrt((dsqr[i] -n*dsum[i]*dsum[i]) /(n-1))
    data = [[(d[i] -dsum[i]) /dsqr[i] for i in dims]
            for d in data]      # z-normalize the data points and

    # --- compute distance matrix ---
    dsqr = dict()               # dictionary for squared distances
    n    = len(data)            # get the number of data points
    pts  = cls = range(n)       # and the range of cluster indices
    for p in pts:               # traverse the data points
        dsqr[p,p] = oo          # self-distance is always zero
        for c in range(p):      # compute the distance matrix
            diff = [a-b for a,b in zip(data[p],data[c])]
            dsqr[p,c] = dsqr[c,p] = sum([d*d for d in diff])

    # --- hierarchical agglomerative clustering ---
    wgts = [1.0 for p in pts]   # initialize the cluster weights
    mrgs = []                   # and the cluster mergers
    while len(cls) > 1:         # while not all clusters are merged
        mx = my = 0; best = oo  # initialize the cluster indices
        for i in range(len(cls)): # traverse the cluster pairs and
            for k in range(i):  # find pair with smallest distance
                x = cls[i]; y = cls[k]
                if dsqr[x,y] < best: best = dsqr[x,y]; mx = x; my = y
        c = len(data)           # get the next cluster id and
        m = (mx,my,c)           # form the merger to execute
        mrgs.append(m)          # store the cluster merger
        wx = wgts[mx]           # get the cluster weights
        wy = wgts[my]; w = wx+wy# and compute their sum
        wx /= w; wy /= w        # compute relative weights
        # print(m,best,w)
        ctr  = [wx*x+wy*y for x,y in zip(data[mx],data[my])]
        data.append(ctr)        # compute new cluster center
        wgts.append(w)          # and the new cluster weight
        cls = [i for i in cls if i != mx and i != my]+[c]
        for p in cls:           # compute distance to new center
            diff = [a-b for a,b in zip(data[p],ctr)]
            dsqr[p,c] = dsqr[c,p] = sum([d*d for d in diff])

    # --- draw diagram ---
    with open(name+'.mp', 'w') as out:
        out.write('u :=  1mm;\n')
        out.write('s := 16mm;\n\n')
        out.write('path p;\n\n')

        #out.write('vardef hue (expr value, expand) =\n')
        #out.write('  save k; numeric k;\n')
        #out.write('  k = (1+2expand)*value+(1-expand);\n')
        #out.write('  if     k < 1: ((k-0) *blue  +(1-k) *black)\n')
        #out.write('  elseif k < 2: ((k-1) *red   +(2-k) *blue)\n')
        #out.write('  else:         ((k-2) *black +(3-k) *red) fi\n')
        #out.write('enddef;\n\n')

        #out.write('vardef hue (expr value, expand) =\n')
        #out.write('  save k; numeric k;\n')
        #out.write('  k = (2+2expand)*value+(1-expand);\n')
        #out.write('  if     k < 1: ((k-0) *black +(1-k) *black)\n')
        #out.write('  elseif k < 2: ((k-1) *blue  +(2-k) *black)\n')
        #out.write('  elseif k < 3: ((k-2) *red   +(3-k) *blue)\n')
        #out.write('  else:         ((k-3) *black +(4-k) *red) fi\n')
        #out.write('enddef;\n\n')

        #out.write('vardef hue (expr value, expand) =\n')
        #out.write('  save k; numeric k;\n')
        #out.write('  k = (2+2expand)*value+(1-expand);\n')
        #out.write('  if     k < 1: ((k-0) *black +(1-k) *black)\n')
        #out.write('  elseif k < 2: ((k-1) *green +(2-k) *black)\n')
        #out.write('  elseif k < 3: ((k-2) *red   +(3-k) *green)\n')
        #out.write('  else:         ((k-3) *black +(4-k) *red) fi\n')
        #out.write('enddef;\n\n')

        #out.write('vardef hue (expr value, expand) =\n')
        #out.write('  save k; numeric k;\n')
        #out.write('  k = (3+2expand)*value+(1-expand);\n')
        #out.write('  if     k < 1: ((k-0) *black +(1-k) *black)\n')
        #out.write('  elseif k < 2: ((k-1) *blue  +(2-k) *black)\n')
        #out.write('  elseif k < 3: ((k-2) *(1,0.8,0) +(3-k) *blue)\n')
        #out.write('  elseif k < 4: ((k-3) *red   +(4-k) *(1,0.7,0))\n')
        #out.write('  else:         ((k-4) *black +(5-k) *red) fi\n')
        #out.write('enddef;\n\n')

        out.write('vardef hue (expr value, expand) =\n')
        out.write('  save k; numeric k;\n')
        out.write('  k = (3+2expand)*value+(1-expand);\n')
        out.write('  if     k < 1: ((k-0) *0.2white  +(1-k) *black)\n')
        out.write('  elseif k < 2: ((k-1) *0.6green  +(2-k) *0.2white)\n')
        out.write('  elseif k < 3: ((k-2) *(1,0.8,0) +(3-k) *0.6green)\n')
        out.write('  elseif k < 4: ((k-3) *red       +(4-k) *(1,0.7,0))\n')
        out.write('  else:         ((k-4) *black     +(5-k) *red) fi\n')
        out.write('enddef;\n\n')

        #out.write('vardef hue (expr value, expand) =\n')
        #out.write('  save k; numeric k;\n')
        #out.write('  k = (4+2expand)*value+(1-expand);\n')
        #out.write('  if     k < 1: (  0,  0,k+0)\n')
        #out.write('  elseif k < 2: (  0,k-1,  1)\n')
        #out.write('  elseif k < 3: (  0,  1,3-k)\n')
        #out.write('  elseif k < 4: (k-3,  1,  0)\n')
        #out.write('  elseif k < 5: (  1,5-k,  0)\n')
        #out.write('  else:         (6-k,  0,  0) fi\n')
        #out.write('enddef;\n\n')

        out.write('vardef dot (expr x,y,c) =\n')
        out.write('  p := fullcircle scaled u shifted (x,y);\n')
        out.write('  fill p withcolor hue(c,0);\n')
        out.write('  draw p withcolor (0.4*hue(c,0))')
        out.write(        ' withpen pencircle scaled 0.33pt;\n')
        out.write('enddef;\n\n')

        out.write('vardef join (expr a,b,d,c) =\n')
        out.write('  draw a--d--b withcolor hue(c,0)'
                 +' withpen pencircle scaled 0.5pt;\n')
        out.write('  dot(xpart d, ypart d, c);\n')
        out.write('enddef;\n\n')

        fno = 0; a = len(atts)
        for x in range(a-1):   # traverse the attribute pairs
            for y in range(x+1,a):
                out.write('beginfig(%d); %%' % fno)
                out.write(' %s:%s\n' % (atts[x],atts[y]))
                #out.write('  pickup pencircle scaled 0.5pt;\n')
                #out.write('  draw unitsquare scaled 2 shifted (-1,-1)'
                #          ' scaled 2s withcolor 0.7white;\n');
                show = [(d[x],d[y]) for d in data]
                for i in range(n):
                    out.write('  dot(%.5fs,%.5fs,0);\n' \
                              % tuple(show[i]))
                m = len(mrgs)
                for i in range(m):
                    mrg  = mrgs[i]
                    src1 = tuple(show[mrg[0]])
                    src2 = tuple(show[mrg[1]])
                    dst  = tuple(show[mrg[2]])
                    c    = float(i+1)/float(m)
                    out.write('  join((%.5fs,%.5fs),' % src1
                                    +'(%.5fs,%.5fs),' % src2
                                    +'(%.5fs,%.5fs),' % dst
                                    +'%.5f);\n' % c)
                out.write('endfig;\n\n')
                fno += 1        # next figure number

        out.write('end.\n')
